package com.library.pdf.capture;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.PaintFlagsDrawFilter;
import android.graphics.Rect;
import android.graphics.RectF;
import android.net.Uri;
import android.util.DisplayMetrics;
import android.util.Log;
import android.util.TypedValue;
import android.view.View;
import android.view.ViewGroup;

import androidx.annotation.NonNull;

import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.uimanager.NativeViewHierarchyManager;
import com.facebook.react.uimanager.UIBlock;
import com.library.pdf.Box;
import com.library.pdf.graph.GraphBox;
import com.library.pdf.hr.HrBox;
import com.library.pdf.pdfbox.PdfBox;

import java.io.File;
import java.io.FileOutputStream;
import java.util.Collections;
import java.util.Set;
import java.util.WeakHashMap;

/**
 * Snapshot utility class allow to screenshot a view.
 */
public class ViewShot implements UIBlock {
    /**
     * Tag fort Class logs.
     */
    private static final String TAG = ViewShot.class.getSimpleName();
    /**
     * Error code that we return to RN.
     */
    public static final String ERROR_UNABLE_TO_SNAPSHOT = "E_UNABLE_TO_SNAPSHOT";
    private final int tag;
    private final double pdfQuality;
    private final File output;
    private final Promise promise;
    private double margin = 0;
    private double realScale;
    private double pageWidth = 0;
    private double pageHeight = 0;

    public ViewShot(int tag, double realScale, ReadableMap size, double pdfQuality, File output, Promise promise) {
        this.pageWidth = size.getDouble("width");
        this.pageHeight = size.getDouble("height");
        this.margin = size.hasKey("margin") ? size.getDouble("margin") : 0;
        this.tag = tag;
        this.realScale = realScale;
        this.pdfQuality = pdfQuality;
        this.output = output;
        this.promise = promise;
    }

    @Override
    public void execute(NativeViewHierarchyManager nativeViewHierarchyManager) {
        final View view = nativeViewHierarchyManager.resolveView(tag);
        if (view == null) {
            Log.e(TAG, "No view found with reactTag: " + tag, new AssertionError());
            promise.reject(ERROR_UNABLE_TO_SNAPSHOT, "No view found with reactTag: " + tag);
            return;
        }

        captureViewImpl(view);
    }

    private PaintFlagsDrawFilter antialiasFilter =
            new PaintFlagsDrawFilter(0, Paint.ANTI_ALIAS_FLAG | Paint.FILTER_BITMAP_FLAG);

    /**
     * Screenshot a view and return the captured bitmap.
     * @param view the view to capture
     */
    private void captureViewImpl(@NonNull final View view) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                int w = (int) (view.getWidth() * realScale);
                int h = (int) (view.getHeight() * realScale);
                if (w <= 0 || h <= 0) {
                    promise.reject(new RuntimeException("Impossible to snapshot the view: view is invalid"));
                    return;
                }
                DisplayMetrics dm = view.getContext().getResources().getDisplayMetrics();
                //int pageW = (int)pageWidth;
                int pageW = (int)TypedValue.applyDimension(TypedValue.COMPLEX_UNIT_DIP, (float) pageWidth, dm);
                //int pageH = (int)pageHeight;
                int pageH = (int) TypedValue.applyDimension(TypedValue.COMPLEX_UNIT_DIP, (float) pageHeight, dm);
                float downscaleY = (float)(pageH - margin * 2) / h;
                float downscaleX = (float)(pageW - margin * 2) / w;
                float downscale = 1;
                if(downscaleX < 1 || downscaleY < 1){
                    downscale = Math.min(downscaleX, downscaleY);
                }
                try {
                    Bitmap background = Bitmap.createBitmap(pageW, pageH, Bitmap.Config.RGB_565);
                    Canvas canvas = new Canvas(background);
                    canvas.drawColor(Color.WHITE);
                    canvas.setDrawFilter(antialiasFilter);
                    Paint paint = new Paint();
                    double offsetX = Math.max(0, (pageW - w * downscale)/2);
                    double offsetY = Math.max(0, (pageH - h * downscale)/2);

                    ViewGroup viewGroup = (ViewGroup) view;
                    for (int i = 0; i < viewGroup.getChildCount(); i++) {
                        View box = viewGroup.getChildAt(i);
                        if (box instanceof PdfBox || box instanceof HrBox || box instanceof GraphBox) {
                            Box viewBox = (Box)box;
                            Bitmap bitmap = viewBox.getBitmap((int) pdfQuality);
                            Rect src = new Rect(0, 0, bitmap.getWidth(), bitmap.getHeight());
                            float left = (float)(offsetX + viewBox.getLeft() * realScale * downscale);
                            float top = (float)(offsetY + viewBox.getTop() * realScale * downscale);
                            float right = (float)(left + viewBox.getWidth()* realScale * downscale);
                            float bottom = (float)(top + viewBox.getHeight() * realScale * downscale);
                            RectF dest = new RectF(left, top, right, bottom);
                            canvas.drawBitmap(bitmap, src, dest, paint);
                            recycleBitmap(bitmap);
                        }
                    }

                    background.compress(Bitmap.CompressFormat.JPEG, 100, new FileOutputStream(output));
                    promise.resolve(Uri.fromFile(output).toString());
                } catch (Exception e){
                    e.printStackTrace();
                    promise.reject(e);
                }
            }
        }).start();
    }

    /**
     * Synchronization guard.
     */
    private static final Object guardBitmaps = new Object();
    /**
     * Reusable bitmaps for screenshots.
     */
    private static final Set<Bitmap> weakBitmaps = Collections.newSetFromMap(new WeakHashMap<Bitmap, Boolean>());

    /**
     * Return bitmap to set of available.
     */
    private static void recycleBitmap(@NonNull final Bitmap bitmap) {
        synchronized (guardBitmaps) {
            weakBitmaps.add(bitmap);
        }
    }
}
