package com.library.pdf.hr;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Color;
import android.util.Log;
import android.view.View;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import static com.facebook.react.bridge.UiThreadUtil.runOnUiThread;

import com.library.pdf.Utils;
public class HrManager {
    public static final String EVENT_HR_LOADED = "topHrLoaded";
    public static final String EVENT_HR_DOWNLOADED = "topHrDownloaded";
    public static final String EVENT_HR_ERROR = "topHrError";
    private static final int MAX_RETRY = 2;
    private static HrManager instance;
    private Scheduler scheduler;

    private HrManager(){
        init();
    }

    public static HrManager getInstance() {
        if (instance == null) {
            synchronized (HrManager.class){
                if (instance == null) {
                    instance = new HrManager();
                }
            }
        }
        return instance;
    }

    public static void clearQueue(){
        if(HrManager.getInstance().scheduler != null) {
            HrManager.getInstance().scheduler.clear();
        }
        HrManager.startQueue();
    }

    public static void setPoolSize(int size){
        if(HrManager.getInstance().scheduler != null) {
            HrManager.getInstance().scheduler.setPoolSize(size);
        }
    }

    public static void startQueue(){
        HrManager.getInstance().init();
    }

    public void init(){
        scheduler = new Scheduler();
    }

    public static void pauseQueue(){
        if(HrManager.getInstance().scheduler != null) {
            HrManager.getInstance().scheduler.pause();
        }
    }

    public static void resumeQueue(){
        if(HrManager.getInstance().scheduler != null) {
            HrManager.getInstance().scheduler.resume();
        }
    }

    void addToList(final HrBox view){
        if(view.isVisible() && view.getRetry() < MAX_RETRY) {
            view.setVisibility(View.VISIBLE);
            try {
                scheduler.schedule(new Runnable() {
                    @Override
                    public void run() {
                        try {
                            renderBox(view);
                        } catch (OutOfMemoryError e) {
                            System.gc();
                            view.setRetry(view.getRetry() + 1);
                            addToList(view);
                        }
                    }
                });
            } catch (OutOfMemoryError ome){
                System.gc();
                view.setRetry(view.getRetry() + 1);
                addToList(view);
            }
        } else if(!view.isVisible()) {
            view.setImageResource(0);
            view.setVisibility(View.GONE);
        }
    }

    public static Bitmap getBitmapFromURL(HrBox view, String destination) throws IOException {
        URL url = new URL(view.getUrl());
        HttpURLConnection connection = (HttpURLConnection) url.openConnection();
        connection.setDoInput(true);
        connection.connect();
        File destinationFile = new File(destination);

        InputStream input = connection.getInputStream();
        OutputStream output = new FileOutputStream(destinationFile);
        byte[] buffer = new byte[4 * 1024];
        int read;
        while ((read = input.read(buffer)) != -1) {
            output.write(buffer, 0, read);
        }
        output.flush();
        return BitmapFactory.decodeFile(destination, getImageConfig(view.isGraph()));
    }

    private static BitmapFactory.Options getImageConfig(boolean isGraph) {
        BitmapFactory.Options localOptions = new BitmapFactory.Options();
        localOptions.inScaled = false;
        localOptions.inDither = false;
        localOptions.inPreferredConfig = isGraph ? Bitmap.Config.RGB_565 : Bitmap.Config.ARGB_8888;
        return localOptions;
    }

    private Bitmap bitmapFromZip(HrBox view) throws IOException, OutOfMemoryError {
        String archive = view.getZipPath();
        String fileName = new File(view.getPath()).getName();
        BufferedInputStream bis;
        Bitmap image = null;
        ZipEntry entry;
        try (ZipFile zipfile = new ZipFile(archive)) {
            entry = zipfile.getEntry(fileName);
            if (entry != null) {
                bis = new BufferedInputStream(zipfile.getInputStream(entry));
                image = BitmapFactory.decodeStream(bis, null, getImageConfig(view.isGraph()));
            }
        }
        return image;
    }

    public Bitmap getBitmap(HrBox view){
        String hrPath = view.getPath();
        String zipPath = view.getZipPath();
        File hrFile = new File(hrPath);
        File zipFile = new File(zipPath);
        Bitmap hrBitmap = null;
        try {
            if (hrFile.exists()) {
                Log.d("HR_MANAGER_FROM_FILE", view.toString());
                hrBitmap = BitmapFactory.decodeFile(hrPath, getImageConfig(view.isGraph()));
                //delete and try to download again, probably is a zombie
                if(hrBitmap == null) {
                    boolean deleted = hrFile.delete();
                    Log.d("HR_MANAGER_DELETE", hrPath + ": " + deleted);
                }
            }

            if(hrBitmap == null) {
                if (zipFile.exists()) {
                    Log.d("HR_MANAGER_FROM_ZIP", view.toString());
                    hrBitmap = bitmapFromZip(view);
                } else if (view.getUrl() != null) {
                    Log.d("HR_MANAGER_FROM_URL", view.toString());
                    hrBitmap = getBitmapFromURL(view, hrPath);
                    if (hrBitmap != null && view.getListener() != null) {
                        view.getListener().onDownloaded(view);
                    }
                }
            }
        } catch (Exception e){
            if (view.getListener() != null) {
                view.getListener().onError(view, e);
            }
            e.printStackTrace();
        }
        return hrBitmap;
    }

    private void renderBox(final HrBox view){
        try {
            Log.d("HR_MANAGER_DRAW", view.toString());
            if(!view.isDrawn() || view.isInvalidate()) {
                view.setInvalidate(false);

                Bitmap hrBitmap = getBitmap(view);
                /*if(!view.isGraph()){
                    hrBitmap = Util.applyCorrectionToBitmap(getBitmap(view));
                }*/
                if (hrBitmap != null) {
                    final Bitmap finalHrBitmap = hrBitmap;
                    runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            if(!view.isSkipEmpty() || !Utils.isEmpty(finalHrBitmap)) {
                            view.setImageBitmap(finalHrBitmap);
                            } else {
                                finalHrBitmap.recycle();
                                Log.d("HR_MANAGER_DRAW", "skip " + view);
                            }
                            if (view.getListener() != null) {
                                view.getListener().onLoadEnd(view);
                            }
                        }
                    });
                }
            }
        } catch (Exception | OutOfMemoryError e){
            if (view.getListener() != null) {
                view.getListener().onError(view, e);
            }
            e.printStackTrace();
        }
    }
}
