package com.library.pdf.pdf;

import android.content.Context;
import android.graphics.Bitmap;
import android.net.Uri;
import android.os.ParcelFileDescriptor;
import android.util.Log;

import androidx.annotation.NonNull;

import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.uimanager.UIManagerModule;
import com.library.pdf.capture.ViewShot;
import com.library.pdfium.PdfDocument;
import com.library.pdfium.PdfiumCore;
import com.library.pdfium.util.Size;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class RCTPdfUtilityModule extends ReactContextBaseJavaModule {
    private ExecutorService executors;


    public RCTPdfUtilityModule(ReactApplicationContext reactContext) {
        super(reactContext);
        executors = Executors.newFixedThreadPool(1);
    }
        
    @Override
    public String getName() {
        return "PdfUtility";
    }

    @ReactMethod
    public void pdfToPng( final ReadableMap config, final Promise promise){
        executors.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    String source = config.getString("source");
                    String destination = config.getString("destination");
                    boolean isDoubleTruck = config.getBoolean("isDoubleTruck");
                    int pngWidthLimit = config.getInt("pngWidthLimit");
                    int pageNumber = 0;
                    File fDestination = new File(destination);
                    File fDestinationProc = new File(destination + ".proc");
                    String oldDestination = destination.replace(fDestination.getName(), "text.png");
                    File fDestinationOld = new File(oldDestination);
                    if (fDestination.exists()) {
                        promise.resolve(destination);
                    } else if(fDestinationOld.exists()){
                        promise.resolve(oldDestination);
                    } else {
                        if(fDestinationProc.exists()){
                            promise.reject(new Exception("Impossible to generate png on this page " + destination));
                        } else {
                            try {
                                fDestinationProc.createNewFile();
                            } catch (IOException ioe){
                                ioe.printStackTrace();
                            }
                            long start = System.currentTimeMillis();
                            Log.d(getName(), "Start to generate png for " + source);
                            Bitmap bmp = generatePngThumbnail(getReactApplicationContext(), source, pageNumber, pngWidthLimit, isDoubleTruck);
                            long middle = System.currentTimeMillis();
                            Log.d(getName(), "generated png for " + source + " in " + (middle - start) / 1000f);
                            if (bmp == null) {
                                promise.reject(new NullPointerException("Bitmap null"));
                            } else {
                                saveImage(bmp, destination);
                                promise.resolve(destination);
                            }
                            fDestinationProc.delete();
                            long end = System.currentTimeMillis();
                            Log.d(getName(), "finish to generate png for " + source + " in " + (end - start) / 1000f);
                        }
                    }
                } catch(Exception | OutOfMemoryError e ) {
                    promise.reject(e);
                }
            }
        });
    }

    public static HashMap<String,Integer> getPageSize(Context context, String source, int pageNumber){
        try {
            HashMap<String,Integer> size = new HashMap<>();
            PdfiumCore pdfiumCore = new PdfiumCore(context);
            ParcelFileDescriptor fd = context.getContentResolver().openFileDescriptor(Uri.fromFile(new File(source)), "r");
            PdfDocument pdfDocument = pdfiumCore.newDocument(fd);
            Size s = pdfiumCore.getPageSize(pdfDocument,0);
            size.put("width", s.getWidth());
            size.put("height", s.getHeight());
            pdfiumCore.closeDocument(pdfDocument);
            return size;
        } catch( Exception | OutOfMemoryError e ){
            return null;
        }
    }

    public static Bitmap generatePngThumbnail(Context context, String source, int pageNumber, int pngWidthLimit, boolean isDoubleTruck){
        try {
            PdfiumCore pdfiumCore = new PdfiumCore(context);
            ParcelFileDescriptor fd = context.getContentResolver().openFileDescriptor(Uri.fromFile(new File(source)), "r");
            PdfDocument pdfDocument = pdfiumCore.newDocument(fd);
            pdfiumCore.openPage(pdfDocument, pageNumber);

            //float scale = isDoubleTruck ? 2 : 2;
            int width = pdfiumCore.getPageWidthPoint(pdfDocument, 0) ;
            int height = pdfiumCore.getPageHeightPoint(pdfDocument, 0);
            int displayWidth = Math.min(context.getResources().getDisplayMetrics().widthPixels, context.getResources().getDisplayMetrics().heightPixels);
            if(isDoubleTruck){
                displayWidth *= 2;
                pngWidthLimit *= 2;
            }

            float scaleX = (float)displayWidth / width;
            int pngWidth = (int) Math.min(pngWidthLimit, displayWidth);
            int pngHeight = (int) (height * scaleX);
            Bitmap bmp = getPageBitmap(pdfiumCore, pdfDocument, pageNumber, pngWidth, pngHeight, pngWidth, pngHeight, 1);
            pdfiumCore.closeDocument(pdfDocument);
            return bmp;
        } catch( Exception | OutOfMemoryError e ){
            return null;
        }
    }


    private static Bitmap getPageBitmap(PdfiumCore pdfiumCore, PdfDocument document, int pageNumber, int width, int height, int pageWidth, int pageHeight, float scale) {
        Bitmap bmp = Bitmap.createBitmap((int) (width*scale), (int) (height*scale), Bitmap.Config.ARGB_8888);
        pdfiumCore.renderPageBitmap(document, bmp, pageNumber, 0, 0, (int)(pageWidth*scale), (int)(pageHeight*scale));
        return bmp;
    }

    @ReactMethod
    public void print( ReadableMap params, Promise promise) {
        try {
            File outputFile = createTempFile(getReactApplicationContext());
            final UIManagerModule uiManager = this.getReactApplicationContext().getNativeModule(UIManagerModule.class);
            int tag = params.getInt("target");
            double scale = params.getDouble("scale");
            ReadableMap size = params.getMap("size");
            double pdfQuality = params.getDouble("pdfQuality");
            uiManager.addUIBlock(new ViewShot(tag, scale, size, pdfQuality, outputFile, promise));
        } catch (Throwable ex) {
            ex.printStackTrace();
            promise.reject("Failed to snapshot view tag ");
        }
    }

    public static void saveImage(Bitmap bmp, String destination ) throws Exception {
        FileOutputStream out = null;
        Exception error = null;
        try {
            File file = new File(destination );
            File folder = new File(file.getParent());
            if (!folder.exists()){
                folder.mkdirs();
            }
            out = new FileOutputStream(file);
            bmp.setHasAlpha(true);
            bmp.setPremultiplied(false);
            bmp.compress(Bitmap.CompressFormat.PNG, 100, out);
        } catch (Exception e) {
            error = e;
        } finally {
            try {
                if (out != null) {
                    out.close();
                }
                if(bmp != null && !bmp.isRecycled()){
                    bmp.recycle();
                }
            } catch (Exception e) {
                error = e;
            }
        }
        if(error != null){
            throw error;
        }
    }

    private static final String TEMP_FILE = "PrintArticleGraph.jpeg";
    /**
     * Create a temporary file in the cache directory on either internal or external storage,
     * whichever is available and has more free space.
     */
    @NonNull
    private File createTempFile(@NonNull final Context context) {
        String cachePath = context.getCacheDir().getPath();
        String file = cachePath + "/" + TEMP_FILE;
        File tempFile = new File(file);
        if(tempFile.exists()){
            tempFile.delete();
        }
        return tempFile;
    }
}
