package com.library.pdf.pdfbox;

import android.content.Context;
import android.graphics.Bitmap;
import android.net.Uri;
import android.os.CountDownTimer;
import android.os.ParcelFileDescriptor;
import android.util.Log;
import android.view.View;

import com.library.pdf.Utils;
import com.library.pdfium.PdfDocument;
import com.library.pdfium.PdfiumCore;

import java.io.File;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class PDFiumManager {
    private static final String TAG = "PDFIUM_MANAGER";
    private static PDFiumManager instance;
    private WeakReference<Context> context;
    private PdfiumCore core;
    private HashMap<String, PdfDocument> documents;
    private HashMap<String,Long> lastDocumentsAccess;
    private ExecutorService executors;
    private int pageNumber;
    private CountDownTimer timer;

    private PDFiumManager(Context context){
        core = new PdfiumCore(context);
        this.context = new WeakReference<>(context);
        executors = Executors.newFixedThreadPool(1);
        pageNumber = 0;
        documents = new HashMap<>();
        lastDocumentsAccess = new HashMap<>();
        timer = new PdfDocumentTimer() {
            @Override
            public void onFinish() {
                try {
                    Set<String> paths = lastDocumentsAccess.keySet();
                    if (core != null) {
                        for (String key : paths) {
                            long unusedForMills = System.currentTimeMillis() - lastDocumentsAccess.get(key);
                            if (unusedForMills > 60000) {
                                Log.d(TAG, "close document: " + key);
                                closeDocument(key);
                                documents.remove(key);
                            }
                        }
                    }
                    timer.start();
                } catch (Exception e){
                    e.printStackTrace();
                }
            }
        }.start();
    }

    public static PDFiumManager getInstance(Context context) {
        if (instance == null) {
            synchronized (PDFiumManager.class){
                if (instance == null) {
                    instance = new PDFiumManager(context);
                }
            }
        }
        return instance;
    }

    private PdfDocument getDocument(String path){
        if(documents.get(path) == null){
            documents.put(path, openDocument(path));
        }
        lastDocumentsAccess.put(path, System.currentTimeMillis());
        Log.d(TAG, "get document: " + path);
        return documents.get(path);
    }

    public PdfiumCore getCore(){
        return core;
    }

    public PdfDocument openDocument(String path){
        PdfDocument document = null;
        try {
            ParcelFileDescriptor fd = getContext().getContentResolver().openFileDescriptor(Uri.fromFile(new File(path)), "r");
            document = core.newDocument(fd);
            core.openPage(document, 0);
            Log.d(TAG, "open document: " + path);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return document;
    }

    private void closeDocument(String path){
        if(core != null && documents.get(path) != null ){
            core.closeDocument(documents.get(path));
        }
    }

    void addToList(final PdfBox view){
        if(view.isVisible()) {
            view.setVisibility(View.VISIBLE);
            executors.submit(new Runnable() {
                @Override
                public void run() {
                    renderBox(view);
                }
            });
        } else {
            view.setImageResource(0);
            view.setVisibility(View.GONE);
        }
    }

    public Bitmap getBitmap(PdfBox view, double scale){
        Log.d("PDF_MANAGER_BOX", view.toString());
        PdfDocument document = getDocument(view.getPath());
        scale = scale * 150 / 72;
        return renderBmp(
                core,
                document,
                -(int)(view.getOriginLeft()*scale),
                -(int)(view.getOriginTop()*scale),
                (int)(view.getPageWidth()*scale),
                (int)(view.getPageHeight()*scale),
                (int)(view.getBoxWidth()*scale),
                (int)(view.getBoxHeight()*scale)
        );
    }

    private void  renderBox(PdfBox view){
        try {
            Log.d("PDF_MANAGER_BOX", view.toString());
            double scale = view.getScale();
            Bitmap bmp = getBitmap(view, scale);
            if(!view.isSkipEmpty() || !Utils.isEmpty(bmp)) {
            view.setBitmap(bmp);
            } else {
                bmp.recycle();
                Log.d("PDF_MANAGER_BOX", "skip " + view);
            }
        } catch (Exception e){
            e.printStackTrace();
        }
    }

    private Bitmap renderBmp(PdfiumCore core, PdfDocument document, int left, int top, int width, int height, int bitmapWidth, int bitmapHeight){
        Bitmap bmp = Bitmap.createBitmap(bitmapWidth, bitmapHeight, Bitmap.Config.ARGB_8888);
        bmp.setDensity(getContext().getResources().getDisplayMetrics().densityDpi);
        core.renderPageBitmap(document, bmp, pageNumber, left, top, width, height);
        return bmp;
    }

    public Bitmap renderBmp(Context context, String path, int left, int top, int pageWidth, int pageHeight, int bitmapWidth, int bitmapHeight, double scale){
        scale = scale * 150 / 72;
        Bitmap bmp = Bitmap.createBitmap((int) (bitmapWidth * scale), (int) (bitmapHeight * scale), Bitmap.Config.ARGB_8888);
        bmp.setDensity(context.getResources().getDisplayMetrics().densityDpi);
        PdfDocument document = getDocument(path);
        core.renderPageBitmap(document, bmp, 0, (int) (-left * scale), (int) (-top * scale), (int) (pageWidth * scale), (int) (pageHeight * scale));
        return bmp;
    }

    private Context getContext() {
        return context.get();
    }

    abstract class PdfDocumentTimer extends CountDownTimer{
        public PdfDocumentTimer() {
            super(30000, 30000);
        }

        @Override
        public void onTick(long millisUntilFinished) {

        }

        @Override
        public void onFinish() {
            this.start();
        }
    }
}
