/*
 * Copyright 2016 Bartosz Schiller
 * <p/>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.library.viewer;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.ColorMatrix;
import android.graphics.ColorMatrixColorFilter;
import android.graphics.Paint;
import android.graphics.PaintFlagsDrawFilter;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.drawable.Drawable;
import android.net.Uri;
import android.os.AsyncTask;
import android.os.HandlerThread;
import android.util.AttributeSet;
import android.util.Log;
import android.widget.RelativeLayout;

import com.library.pdfium.PdfiumCore;
import com.library.pdfium.util.Size;
import com.library.pdfium.util.SizeF;
import com.library.viewer.exception.PageRenderingException;
import com.library.viewer.link.DefaultLinkHandler;
import com.library.viewer.link.LinkHandler;
import com.library.viewer.listener.Callbacks;
import com.library.viewer.listener.OnDrawListener;
import com.library.viewer.listener.OnErrorListener;
import com.library.viewer.listener.OnLoadCompleteListener;
import com.library.viewer.listener.OnPageErrorListener;
import com.library.viewer.listener.OnRenderListener;
import com.library.viewer.listener.OnRenderingHandlerListener;
import com.library.viewer.model.PagePart;
import com.library.viewer.source.AssetSource;
import com.library.viewer.source.DocumentSource;
import com.library.viewer.source.UriSource;
import com.library.viewer.util.FitPolicy;
import com.library.viewer.util.MathUtils;
import com.library.viewer.util.Util;

/**
 * It supports animations, zoom, cache, and swipe.
 * <p>
 * To fully understand this class you must know its principles :
 * - The PDF document is seen as if we always want to draw all the pages.
 * - The thing is that we only draw the visible parts.
 * - All parts are the same size, this is because we can't interrupt a native page rendering,
 * so we need these renderings to be as fast as possible, and be able to interrupt them
 * as soon as we can.
 * - The parts are loaded when the current offset or the current zoom level changes
 * <p>
 * Important :
 * - DocumentPage = A page of the PDF document.
 * - UserPage = A page as defined by the user.
 * By default, they're the same. But the user can change the pages order
 * using {@link #load(DocumentSource, String, int[])}. In this
 * particular case, a userPage of 5 can refer to a documentPage of 17.
 */
public class PDFView extends RelativeLayout implements OnRenderingHandlerListener {

    private static final int PDF_LOAD_RETRY = 3;
    private static final String TAG = PDFView.class.getSimpleName();

    /** Rendered parts go to the cache manager */
    CacheManager cacheManager;

    PdfFile pdfFile;

    /**
     * If you picture all the pages side by side in their optimal width,
     * and taking into account the zoom level, the current offset is the
     * position of the left border of the screen in this big picture
     */
    private float currentXOffset = 0;

    /**
     * If you picture all the pages side by side in their optimal width,
     * and taking into account the zoom level, the current offset is the
     * position of the left border of the screen in this big picture
     */
    private float currentYOffset = 0;

    /** The zoom level, always >= 1 */
    private float zoom = 1f;
    private float scale = 1f;
    /** True if the PDFView has been recycled */
    private boolean recycled = true;

    /** Current state of the view */
    public State state = State.DEFAULT;

    /** Async task used during the loading phase to decode a PDF document */
    private DecodingAsyncTask decodingAsyncTask;

    /** The thread {@link #renderingHandler} will run on */
    private final HandlerThread renderingHandlerThread;
    /** Handler always waiting in the background and rendering tasks */
    RenderingHandler renderingHandler;

    private PagesLoader pagesLoader;

    Callbacks callbacks = new Callbacks();

    /** Paint object for drawing */
    private Paint paint;

    /** Policy for fitting pages to screen */
    private FitPolicy pageFitPolicy = FitPolicy.WIDTH;

    /** True if should scroll through pages vertically instead of horizontally */
    private boolean swipeVertical = true;

    private boolean nightMode = false;

    /** Pdfium core for loading and rendering PDFs */
    private PdfiumCore pdfiumCore;

    /**
     * True if bitmap should use ARGB_8888 format and take more memory
     * False if bitmap should be compressed by using RGB_565 format and take less memory
     */
    private boolean bestQuality = false;

    /**
     * True if annotations should be rendered
     * False otherwise
     */
    private boolean annotationRendering = false;

    /** Antialiasing and bitmap filtering */
    private boolean enableAntialiasing = true;
    private PaintFlagsDrawFilter antialiasFilter =
            new PaintFlagsDrawFilter(0, Paint.ANTI_ALIAS_FLAG | Paint.FILTER_BITMAP_FLAG);

    /** Spacing between pages, in px */
    private int spacingPx = 0;

    /** Add dynamic spacing to fit each page separately on the screen. */
    private boolean autoSpacing = false;

    private boolean thumbnailClear = false;

    private boolean doubleTruck = false;

    private int thumbnailMaxWidth = 1080;

    /** Holds info whether view has been added to layout and has width and height */
    private boolean hasSize = false;

    /** Holds last used Configurator that should be loaded when view has size */
    private Configurator waitingDocumentConfigurator;
    private boolean rendering = false;
    private Rect lastThumbnailRectCleared;
    private float thumbnailRatio = 1;
    private boolean attacched;
    private boolean blackAndWhite = false;

    /** Construct the initial view */
    public PDFView(Context context, AttributeSet set) {
        super(context, set);

        renderingHandlerThread = new HandlerThread("PDF renderer");

        if (isInEditMode()) {
            return;
        }

        cacheManager = new CacheManager();
        pagesLoader = new PagesLoader(this);

        paint = new Paint();
        pdfiumCore = new PdfiumCore(context);
        setWillNotDraw(false);
    }

    private void load(DocumentSource docSource, String password) {
        load(docSource, password, null);
    }

    private void load(DocumentSource docSource, String password, int[] userPages) {

        if (!recycled) {
            throw new IllegalStateException("Don't call load on a PDF View without recycling it first.");
        }

        recycled = false;
        // Start decoding document
        decodingAsyncTask = new DecodingAsyncTask(docSource, password, this, pdfiumCore);
        decodingAsyncTask.executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
    }

    /**
     * Go to the given page.
     *
     */
    public void jumpTo() {
        if (pdfFile == null) {
            return;
        }

        moveTo(currentXOffset * zoom, currentYOffset * zoom);
        if (recycled) {
            return;
        }
        loadPages();
    }

    /**
     * Get current position as ratio of document length to visible area.
     * 0 means that document start is visible, 1 that document end is visible
     *
     * @return offset between 0 and 1
     */
    public float getPositionOffset() {
        try {
            float offset;
            if (swipeVertical) {
                offset = -currentYOffset / (pdfFile.getDocLen(zoom) - getHeight());
            } else {
                offset = -currentXOffset / (pdfFile.getDocLen(zoom) - getWidth());
            }
            return MathUtils.limit(offset, 0, 1);
        } catch ( Exception e){
            return 0;
        }
    }

    public void setNightMode(boolean nightMode) {
        this.nightMode = nightMode;
        if (nightMode) {
            ColorMatrix colorMatrixInverted =
                    new ColorMatrix(new float[]{
                            -1, 0, 0, 0, 255,
                            0, -1, 0, 0, 255,
                            0, 0, -1, 0, 255,
                            0, 0, 0, 1, 0});

            ColorMatrixColorFilter filter = new ColorMatrixColorFilter(colorMatrixInverted);
            paint.setColorFilter(filter);
        } else {
            paint.setColorFilter(null);
        }
    }

    void onPageError(PageRenderingException ex) {
        if (!callbacks.callOnPageError(ex.getPage(), ex.getCause())) {
            Log.e(TAG, "Cannot open page " + ex.getPage(), ex.getCause());
        }
    }

    public void recycle() {
        waitingDocumentConfigurator = null;

        // Stop tasks
        if (renderingHandler != null) {
            renderingHandler.stop();
            renderingHandler.removeMessages(RenderingHandler.MSG_RENDER_TASK);
        }
        if (decodingAsyncTask != null) {
            decodingAsyncTask.cancel(true);
        }

        // Clear caches
        cacheManager.recycle();

        if (pdfFile != null) {
            pdfFile.dispose();
            pdfFile = null;
        }

        renderingHandler = null;
        currentXOffset = currentYOffset = 0;
        zoom = 1f;
        recycled = true;
        callbacks = new Callbacks();
        state = State.DEFAULT;
    }

    @Override
    protected void onAttachedToWindow() {
        super.onAttachedToWindow();
        attacched = true;
    }

    @Override
    protected void onDetachedFromWindow() {
        //recycle();
        super.onDetachedFromWindow();
        attacched = false;
    }

    @Override
    protected void onSizeChanged(int w, int h, int oldw, int oldh) {
        hasSize = true;
        if (waitingDocumentConfigurator != null) {
            waitingDocumentConfigurator.load();
        }
        if (isInEditMode() || state != State.SHOWN) {
            return;
        }
        pdfFile.recalculatePageSizes(new Size(w, h));
        if (swipeVertical) {
            moveTo(currentXOffset, -pdfFile.getPageOffset(0, zoom));
        } else {
            moveTo(-pdfFile.getPageOffset(0, zoom), currentYOffset);
        }
        //loadPageByOffset();
    }

    @Override
    protected void onDraw(Canvas canvas) {
        if (isInEditMode()) {
            return;
        }
        if (enableAntialiasing) {
            canvas.setDrawFilter(antialiasFilter);
        }

        Drawable bg = getBackground();
        if (bg == null) {
            canvas.drawColor(nightMode ? Color.BLACK : Color.TRANSPARENT);
        } else {
            bg.draw(canvas);
        }

        if (recycled) {
            return;
        }

        if (state != State.SHOWN) {
            return;
        }

        Log.d(TAG, "start to draw: " + rendering);
        Log.d(TAG, "max cache order: " + pagesLoader.getCacheOrder());
        int partsToDraw = pagesLoader.getCacheOrder();
        // Draws thumbnails
        for (PagePart part : cacheManager.getThumbnails()) {
            drawPart(canvas, part, true, !rendering || partsToDraw == 1 );
            if(partsToDraw == 1){
                drawWithListener(canvas, 0, callbacks.getOnDraw());
            }
        }

         // Draws parts
        Log.d(TAG, "parts: " + cacheManager.toString());
        for (PagePart part : cacheManager.getPageParts()) {
            drawPart(canvas, part, false, false);
            pagesLoader.setCellRendered(part.getCacheOrder());
            int cellNotRendered = pagesLoader.allCellsRendered();
            if(cellNotRendered == -1 && rendering){
                rendering = false;
                Log.d(TAG, "clear thumbnail");
                redraw();
            } else {
                Log.d(TAG, "cache manager parts: " + cacheManager.getPageParts().size());
                Log.d(TAG, "first part not rendered: " + cellNotRendered);
            }
        }
    }

    private void drawWithListener(Canvas canvas, int page, OnDrawListener listener) {
        if (listener != null) {
            SizeF size = pdfFile.getPageSize(page);
            listener.onLayerDrawn(canvas, size.getWidth(), size.getHeight(), page);
        }
    }

    /** Draw a given PagePart on the canvas */
    private void drawPart(Canvas canvas, PagePart part, boolean isThumbnail, boolean clearNewArea) {
        // Can seem strange, but avoid lot of calls
        RectF pageRelativeBounds = part.getPageRelativeBounds();
        Bitmap renderedBitmap = part.getRenderedBitmap();
        if (renderedBitmap.isRecycled()) {
            return;
        }

        if(isThumbnail) {
            if(getPagesLoader().mustClearVisibleAreaOnThumb()) {
                if (clearNewArea) {
                    lastThumbnailRectCleared = pagesLoader.getThumbnailAreaToClear(0);
                    renderedBitmap = Util.mergeBitmaps(renderedBitmap, lastThumbnailRectCleared, 0);
                    cacheManager.clearPassiveCache();
                } else {
                    renderedBitmap = Util.mergeBitmaps(renderedBitmap, lastThumbnailRectCleared, 0);
                }
            } else {
                cacheManager.clearPassiveCache();
            }
        }
        if(renderedBitmap != null) {
            SizeF size = pdfFile.getPageSize(part.getPage());
            Rect srcRect = new Rect(0, 0, renderedBitmap.getWidth(), renderedBitmap.getHeight());
            float offsetX = pageRelativeBounds.left * size.getWidth();
            float offsetY = pageRelativeBounds.top * size.getHeight();
            float width = pageRelativeBounds.width() * size.getWidth();
            float height = pageRelativeBounds.height() * size.getHeight();

            // If we use float values for this rectangle, there will be
            // a possible gap between page parts, especially when
            // the zoom level is high.
            RectF dstRect = new RectF((int) offsetX, (int) offsetY, (int) (offsetX + width), (int) (offsetY + height));
            // Check if bitmap is in the screen
            canvas.drawBitmap(renderedBitmap, srcRect, dstRect, paint);
        }
    }



    /**
     * Load all the parts around the center of the screen,
     * taking into account X and Y offsets, zoom level, and
     * the current page displayed
     */
    public void loadPages() {
        Log.d(TAG, "load pages");
        if (pdfFile == null || renderingHandler == null) {
            return;
        }

        // Cancel all current tasks
        renderingHandler.removeMessages(RenderingHandler.MSG_RENDER_TASK);
        cacheManager.makeANewSet();

        pagesLoader.loadPages();
        rendering = true;
        redraw();
    }

    /** Called when the PDF is loaded */
    void loadComplete(PdfFile pdfFile, int retry) {
        state = State.LOADED;

        this.pdfFile = pdfFile;

        try {
            if (!renderingHandlerThread.isAlive()) {
                renderingHandlerThread.start();
            }

            renderingHandler = new RenderingHandler(renderingHandlerThread.getLooper(), this);
            renderingHandler.setListener(this);
            renderingHandler.start();
            callbacks.callOnLoadComplete(pdfFile.getPagesCount());

            jumpTo();
        } catch (OutOfMemoryError e){
            e.printStackTrace();
            System.gc();
            if(retry < PDF_LOAD_RETRY){
                loadComplete(pdfFile, retry + 1);
            }
        }
    }

    private boolean postAction(Runnable runnable){
        return attacched && post(runnable);
    }


    @Override
    public void onTaskBitmapRendered(final PagePart part) {
        Log.d(TAG, "part rendered: " + part.getCacheOrder());
        if (!postAction(new Runnable() {
            @Override
            public void run() {
                onBitmapRendered(part);
            }
        })) {
            try {
                onBitmapRendered(part);
            } catch (Exception ex){
                ex.printStackTrace();
            }
        }
    }

    @Override
    public void onTaskBitmapRecycle(final PagePart part) {
        Log.d(TAG, "part recycled: " + part.getCacheOrder());
        if (!postAction(new Runnable() {
            @Override
            public void run() {
                part.getRenderedBitmap().recycle();
            }
        })) {
            try {
                part.getRenderedBitmap().recycle();
            } catch (Exception ex){
                ex.printStackTrace();
            }
        }
    }

    @Override
    public void onTaskBitmapFailed(final RenderingHandler.RenderingTask task) {
        Log.d(TAG, "part is null: " + task.cacheOrder);
        if (!postAction(new Runnable() {
            @Override
            public void run() {
                onBitmapFailed(task);
            }
        })) {
            try {
                onBitmapFailed(task);
            } catch (Exception ex){
                ex.printStackTrace();
            }
        }
    }

    @Override
    public void onTaskPageError(final PageRenderingException e) {
        if (!postAction(new Runnable() {
            @Override
            public void run() {
                onPageError(e);
            }
        })) {
            try {
                onPageError(e);
            } catch (Exception ex){
                ex.printStackTrace();
            }
        }
    }

    void loadError(Throwable t) {
        state = State.ERROR;
        // store reference, because callbacks will be cleared in recycle() method
        OnErrorListener onErrorListener = callbacks.getOnError();
        recycle();
        invalidate();
        if (onErrorListener != null) {
            onErrorListener.onError(t);
        } else {
            Log.e("PDFView", "load pdf error", t);
        }
    }

    protected void redraw() {
        invalidate();
    }

    public synchronized void onBitmapFailed(RenderingHandler.RenderingTask task){
        pagesLoader.removeCells(task.cacheOrder);
        redraw();
    }

    /**
     * Called when a rendering task is over and
     * a PagePart has been freshly created.
     *
     * @param part The created PagePart.
     */
    public synchronized void onBitmapRendered(PagePart part) {
        // when it is first rendered part
        if (state == State.LOADED) {
            state = State.SHOWN;
            callbacks.callOnRender(pdfFile.getPagesCount());
        }

        if (part.isThumbnail()) {
            cacheManager.cacheThumbnail(part);
        } else {
            cacheManager.cachePart(part);
        }
        redraw();
    }

    /**
     * Move to the given X and Y offsets, but check them ahead of time
     * to be sure not to go outside the the big strip.
     *
     * @param offsetX    The big strip X offset to use as the left border of the screen.
     * @param offsetY    The big strip Y offset to use as the right border of the screen.
     */
    public void moveTo(float offsetX, float offsetY) {
        currentXOffset = offsetX;
        currentYOffset = offsetY;
        float positionOffset = getPositionOffset();
        callbacks.callOnPageScroll(0, positionOffset);
        redraw();
    }

    public float getScale(){
        return scale;
    }

    public void setScale(float scale){
        this.scale = scale;
    }

    public void useBestQuality(boolean bestQuality) {
        this.bestQuality = bestQuality;
    }

    public boolean isBestQuality() {
        return bestQuality;
    }

    public boolean isSwipeVertical() {
        return swipeVertical;
    }

    private void setSwipeVertical(boolean swipeVertical) {
        this.swipeVertical = swipeVertical;
    }

    public void enableAnnotationRendering(boolean annotationRendering) {
        this.annotationRendering = annotationRendering;
    }

    public boolean isAnnotationRendering() {
        return annotationRendering;
    }

    public void enableAntialiasing(boolean enableAntialiasing) {
        this.enableAntialiasing = enableAntialiasing;
    }

    public int getSpacingPx() {
        return spacingPx;
    }

    public boolean doAutoSpacing() {
        return autoSpacing;
    }

    private void setSpacing(int spacing) {
        this.spacingPx = Util.getDP(getContext(), spacing);
    }

    private void setAutoSpacing(boolean autoSpacing) {
        this.autoSpacing = autoSpacing;
    }

    private void setPageFitPolicy(FitPolicy pageFitPolicy) {
        this.pageFitPolicy = pageFitPolicy;
    }

    public FitPolicy getPageFitPolicy() {
        return pageFitPolicy;
    }

    /** Use URI as the pdf source, for use with content providers */
    public Configurator fromUri(Uri uri) {
        return new Configurator(new UriSource(uri));
    }

    /** Use an asset file as the pdf source */
    public Configurator fromAsset(String assetName) {
        return new Configurator(new AssetSource(assetName));
    }

    public void setDoubleTruck(boolean doubleTruck) {
        this.doubleTruck = doubleTruck;
    }

    public boolean isDoubleTruck(){
        return this.doubleTruck;
    }

    public void setThumbnailMaxWidth(int thumbnailMaxWidth) {
        this.thumbnailMaxWidth = thumbnailMaxWidth;
    }

    public int getThumbnailMaxWidth(){
        return this.thumbnailMaxWidth;
    }

    public void setThumbnailClear(boolean thumbnailClear) {
        this.thumbnailClear = thumbnailClear;
    }

    public boolean isThumbnailClear(){
        return this.thumbnailClear;
    }

    public void setThumbnailRatio(float thumbnailRatio) {
        this.thumbnailRatio = thumbnailRatio;
    }

    public void setBlackAndWhite(boolean blackAndWhite){
        this.blackAndWhite = blackAndWhite;
    }

    public boolean isBlackAndWhite(){
        return blackAndWhite;
    }

    public float getThumbnailRatio(){
        return this.thumbnailRatio;
    }

    public enum State {DEFAULT, LOADED, SHOWN, ERROR}

    public class Configurator {

        private final DocumentSource documentSource;

        private int[] pageNumbers = null;

        private boolean enableSwipe = true;

        private OnDrawListener onDrawListener;

        private OnLoadCompleteListener onLoadCompleteListener;

        private OnErrorListener onErrorListener;

        private OnRenderListener onRenderListener;

        private OnPageErrorListener onPageErrorListener;

        private LinkHandler linkHandler = new DefaultLinkHandler(PDFView.this);

        private boolean swipeHorizontal = false;

        private boolean annotationRendering = true;

        private String password = null;

        private boolean antialiasing = true;

        private int spacing = 0;

        private boolean autoSpacing = false;

        private FitPolicy pageFitPolicy = FitPolicy.WIDTH;

        private boolean nightMode = false;

        private Configurator(DocumentSource documentSource) {
            this.documentSource = documentSource;
        }

        public Configurator onDraw(OnDrawListener onDrawListener) {
            this.onDrawListener = onDrawListener;
            return this;
        }

        public Configurator onLoad(OnLoadCompleteListener onLoadCompleteListener) {
            this.onLoadCompleteListener = onLoadCompleteListener;
            return this;
        }

        public Configurator onError(OnErrorListener onErrorListener) {
            this.onErrorListener = onErrorListener;
            return this;
        }

        public Configurator onPageError(OnPageErrorListener onPageErrorListener) {
            this.onPageErrorListener = onPageErrorListener;
            return this;
        }

        public Configurator onRender(OnRenderListener onRenderListener) {
            this.onRenderListener = onRenderListener;
            return this;
        }

        public Configurator swipeHorizontal(boolean swipeHorizontal) {
            this.swipeHorizontal = swipeHorizontal;
            return this;
        }

        public Configurator password(String password) {
            this.password = password;
            return this;
        }

        public Configurator enableAntialiasing(boolean antialiasing) {
            this.antialiasing = antialiasing;
            return this;
        }

        public Configurator enableAnnotationRendering(boolean annotationRendering){
            this.annotationRendering = annotationRendering;
            return this;
        }

        public Configurator spacing(int spacing) {
            this.spacing = spacing;
            return this;
        }

        public Configurator autoSpacing(boolean autoSpacing) {
            this.autoSpacing = autoSpacing;
            return this;
        }

        public Configurator pageFitPolicy(FitPolicy pageFitPolicy) {
            this.pageFitPolicy = pageFitPolicy;
            return this;
        }

        public void load() {
            if (!hasSize) {
                waitingDocumentConfigurator = this;
                return;
            }
            Log.d("PDFVIEW", "Load: " + this.toString());
            PDFView.this.recycle();
            PDFView.this.callbacks.setOnLoadComplete(onLoadCompleteListener);
            PDFView.this.callbacks.setOnError(onErrorListener);
            PDFView.this.callbacks.setOnDraw(onDrawListener);
            PDFView.this.callbacks.setOnRender(onRenderListener);
            PDFView.this.callbacks.setOnPageError(onPageErrorListener);
            PDFView.this.callbacks.setLinkHandler(linkHandler);
            PDFView.this.setNightMode(nightMode);
            PDFView.this.setSwipeVertical(!swipeHorizontal);
            PDFView.this.enableAnnotationRendering(annotationRendering);
            PDFView.this.enableAntialiasing(antialiasing);
            PDFView.this.setSpacing(spacing);
            PDFView.this.setAutoSpacing(autoSpacing);
            PDFView.this.setPageFitPolicy(pageFitPolicy);

            if (pageNumbers != null) {
                PDFView.this.load(documentSource, password, pageNumbers);
            } else {
                PDFView.this.load(documentSource, password);
            }
        }
    }

    public PagesLoader getPagesLoader() {
        return pagesLoader;
    }
}
