/*
 * Copyright 2017 Bartosz Schiller
 * <p/>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.library.viewer;

import android.graphics.Rect;
import android.graphics.RectF;
import android.util.Log;

import com.library.pdfium.util.SizeF;
import com.library.viewer.exception.PageRenderingException;
import com.library.viewer.util.Constants;
import com.library.viewer.util.MathUtils;
import com.library.viewer.util.Util;

import java.util.HashMap;

public class PagesLoader {

    static final String TAG = "PAGES_LOADER";

    private PDFView pdfView;
    private int cacheOrder;
    private float pageRelativePartWidth;
    private float pageRelativePartHeight;
    private float partRenderWidth;
    private float partRenderHeight;
    private final RectF thumbnailRect = new RectF(0, 0, 1, 1);
    private final int preloadOffset;
    private final Holder holder = new Holder();
    private final GridSize firstGrid = new GridSize();
    private Rect visible;
    private float scale;
    private RectF visibleWithOffset;
    private float colWidth = 0;
    private float rowHeight = 0;
    private HashMap<Integer, Boolean> cellsRendered;

    private class Holder {
        int page;
        int firstRow;
        int lastRow;
        int firstCol;
        int lastCol;
    }

    private class GridSize {
        int rows;
        int cols;
    }

    PagesLoader(PDFView pdfView) {
        this.pdfView = pdfView;
        this.preloadOffset = Util.getDP(pdfView.getContext(), Constants.PRELOAD_OFFSET);
    }

    private void getPageColsRows(GridSize grid, int pageIndex) {
        SizeF size = pdfView.pdfFile.getPageSize(pageIndex);
        grid.rows = MathUtils.ceil(size.getHeight() * scale / Constants.PART_SIZE);
        grid.cols = MathUtils.ceil(size.getWidth() * scale / Constants.PART_SIZE);
    }

    private Holder getPageAndCoordsByOffset(Holder holder, GridSize grid) {
        holder.page = 0;
        getPageColsRows(grid, holder.page);
        SizeF scaledPageSize = pdfView.pdfFile.getScaledPageSize(holder.page, pdfView.getScale());

        rowHeight = scaledPageSize.getHeight() / grid.rows;
        colWidth = scaledPageSize.getWidth() / grid.cols;

        visibleWithOffset = new RectF();
        visibleWithOffset.left = Math.max(visible.left - preloadOffset, 0);
        visibleWithOffset.top = Math.max(visible.top - preloadOffset, 0);
        visibleWithOffset.right = Math.min(visible.right + preloadOffset, scaledPageSize.getWidth());
        visibleWithOffset.bottom = Math.min(visible.bottom + preloadOffset, scaledPageSize.getHeight());
        holder.firstRow = MathUtils.floor(visibleWithOffset.top / rowHeight);
        holder.firstCol = MathUtils.floor(visibleWithOffset.left/ colWidth);
        holder.lastRow = MathUtils.ceil(visibleWithOffset.bottom / rowHeight);
        holder.lastCol = MathUtils.ceil(visibleWithOffset.right/ colWidth);
        return holder;
    }

    private void calculatePartSize(GridSize grid) {
        pageRelativePartWidth = 1f / (float) grid.cols;
        pageRelativePartHeight = 1f / (float) grid.rows;
        partRenderWidth = Constants.PART_SIZE / pageRelativePartWidth;
        partRenderHeight = Constants.PART_SIZE / pageRelativePartHeight;
    }

    private void loadVisible() {
        getPageAndCoordsByOffset(holder, firstGrid);
        try {
            pdfView.pdfFile.openPage(0);
        } catch (PageRenderingException e) {
            e.printStackTrace();
        }

        loadThumbnail(holder.page);

        if(this.mustClearVisibleAreaOnThumb()) {
            loadPageCenter(holder, firstGrid, Constants.CACHE_SIZE);
        }
    }

    public boolean mustClearVisibleAreaOnThumb(){
        Log.d(TAG, "mustClearVisibleAreaOnThumb scale: " + scale);
        return pdfView.isThumbnailClear();
    }

    /**
     * When only part of one page is visible
     *
     * @return loaded parts count
     */
    private int loadPageCenter(Holder firstHolder, GridSize grid, int nbOfPartsLoadable) {
        calculatePartSize(grid);
        return loadPage(firstHolder.page, firstHolder.firstRow, firstHolder.lastRow, firstHolder.firstCol, firstHolder.lastCol, nbOfPartsLoadable);
    }

    private int loadPage(int page, int firstRow, int lastRow, int firstCol, int lastCol, int nbOfPartsLoadable) {
        int loaded = 0;
        for (int row = firstRow; row <= lastRow; row++) {
            for (int col = firstCol; col <= lastCol; col++) {
                if (loadCell(page, row, col, pageRelativePartWidth, pageRelativePartHeight)) {
                    loaded++;
                }
                if (loaded >= nbOfPartsLoadable) {
                    return loaded;
                }
            }
        }
        return loaded;
    }

    private boolean loadCell(int page, int row, int col, float pageRelativePartWidth, float pageRelativePartHeight) {

        float relX = pageRelativePartWidth * col;
        float relY = pageRelativePartHeight * row;
        float relWidth = pageRelativePartWidth;
        float relHeight = pageRelativePartHeight;

        float renderWidth = partRenderWidth;
        float renderHeight = partRenderHeight;
        if (relX + relWidth > 1) {
            relWidth = 1 - relX;
        }
        if (relY + relHeight > 1) {
            relHeight = 1 - relY;
        }
        renderWidth *= relWidth;
        renderHeight *= relHeight;
        RectF pageRelativeBounds = new RectF(relX, relY, relX + relWidth, relY + relHeight);

        if (renderWidth > 0 && renderHeight > 0) {
            if (!pdfView.cacheManager.upPartIfContained(page, pageRelativeBounds, cacheOrder)) {
                Log.d(TAG, "cell: " + cacheOrder + " is a new box");
                try {
                    pdfView.renderingHandler.addRenderingTask(page, renderWidth, renderHeight, pageRelativeBounds, false, cacheOrder, pdfView.isBestQuality(), pdfView.isAnnotationRendering());
                } catch (NullPointerException npe){
                    npe.printStackTrace();
                }
            } else {
                Log.d(TAG, "cell: " + cacheOrder + " is already in cache");
            }
            cellsRendered.put(cacheOrder, false);
            cacheOrder++;
            return true;
        }
        return false;
    }

    private void loadThumbnail(int page) {
        float[] thumbnailSize = getThumbnailSize(page);
        if (!pdfView.cacheManager.containsThumbnail(page, thumbnailRect)) {
            try {
                pdfView.renderingHandler.addRenderingTask(page, thumbnailSize[0], thumbnailSize[1], thumbnailRect, true, 0, pdfView.isBestQuality(), pdfView.isAnnotationRendering());
            } catch (NullPointerException npe){
                npe.printStackTrace();
            }
        }
    }

    private float[] getThumbnailSize(int page){
        int pngWidthLimit = pdfView.getThumbnailMaxWidth();
        if(pdfView.isDoubleTruck()){
            pngWidthLimit *= 2;
        }
        //pdfWidth has correct size and we don't need to multiply by 2 if is double truck
        int pdfWidth = pdfView.pdfFile.getPageWidthPoint(page);
        int pdfHeight = pdfView.pdfFile.getPageHeightPoint(page);
        int width = pngWidthLimit;
        if(pdfView.getWidth() > 0){
            width = pdfView.getWidth();
        }
        if(width < pngWidthLimit / 2){
            width = pngWidthLimit / 2;
        }
        width = (int) (width * pdfView.getThumbnailRatio());
        if(width > pngWidthLimit){
            width = pngWidthLimit;
        }
        int height = (int) (width * pdfHeight / pdfWidth);

        return new float[] { width, height };
    }

    void loadPages() {
        cacheOrder = 1;
        cellsRendered = new HashMap<>();
        scale = pdfView.getScale();
        loadVisible();
    }

    public void setLoaded(Rect newVisibleRect) {
        visible = newVisibleRect;
    }

    public Rect getLoaded() {
        if(visible != null){
            Rect rectWithOffset = new Rect();
            rectWithOffset.left = visible.left - preloadOffset;
            rectWithOffset.right = visible.right + preloadOffset;
            rectWithOffset.top = visible.top - preloadOffset;
            rectWithOffset.bottom = visible.bottom + preloadOffset;
            return rectWithOffset;
        }
        return visible;
    }

    public Rect getThumbnailAreaToClear(int page){
        float[] size = getThumbnailSize(page);
        SizeF scaledPageSize = pdfView.pdfFile.getScaledPageSize(holder.page, pdfView.getScale());
        float scaleX = size[0] / scaledPageSize.getWidth();
        float scaleY = size[1] / scaledPageSize.getHeight();
        Rect loaded = new Rect();
        loaded.left = (int) (MathUtils.floor(visibleWithOffset.left/ colWidth) * colWidth * scaleX);
        loaded.right = (int) ((MathUtils.ceil(visibleWithOffset.right/ colWidth) * colWidth + colWidth) * scaleX);
        loaded.top = (int) (MathUtils.floor(visibleWithOffset.top / rowHeight) * rowHeight * scaleY);
        loaded.bottom = (int) ((MathUtils.ceil(visibleWithOffset.bottom / rowHeight) * rowHeight + rowHeight) * scaleY);
        return loaded;
    }

    public float getLoadedScale(){
        return scale;
    }

    public int getCacheOrder() {
        return cacheOrder;
    }

    public void setCellRendered(int cell){
        if(this.cellsRendered != null && this.cellsRendered.containsKey(cell)){
            this.cellsRendered.put(cell, true);
        }
    }

    public void removeCells(int cell) {
        if(this.cellsRendered != null && this.cellsRendered.containsKey(cell)){
            this.cellsRendered.remove(cell);
        }
        cacheOrder--;
    }


    public int allCellsRendered(){
        if(this.cellsRendered != null){
            Log.d(TAG, "cellsRendered size: " + this.cellsRendered.size());
            Log.d(TAG, "cacheOrder size: " + cacheOrder);
            Log.d(TAG, "cells state: " + this.cellsRendered.toString());
            if(this.cellsRendered.size() == cacheOrder - 1){
                for(Integer cell : cellsRendered.keySet()){
                    if(cellsRendered.get(cell) != Boolean.TRUE){
                        return cell;
                    }
                }
                return -1;
            }
        }
        return 0;
    }
}
