/*
 * Copyright 2016 Bartosz Schiller
 * <p/>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.library.viewer;

import android.graphics.Bitmap;
import android.graphics.Matrix;
import android.graphics.Rect;
import android.graphics.RectF;
import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import android.util.Log;

import com.library.viewer.exception.PageRenderingException;
import com.library.viewer.listener.OnRenderingHandlerListener;
import com.library.viewer.model.PagePart;

/**
 * A {@link Handler} that will process incoming {@link RenderingTask} messages
 * PDF is ready to render.
 */
public class RenderingHandler extends Handler {
    /**
     * {@link Message#what} kind of message this handler processes.
     */
    static final int MSG_RENDER_TASK = 1;
    static final int MAX_RETRY = 1;

    private static final String TAG = RenderingHandler.class.getName();

    private PDFView pdfView;

    private RectF renderBounds = new RectF();
    private Rect roundedRenderBounds = new Rect();
    private Matrix renderMatrix = new Matrix();
    private boolean running = false;
    private OnRenderingHandlerListener listener;

    RenderingHandler(Looper looper, PDFView pdfView) {
        super(looper);
        this.pdfView = pdfView;
    }

    void addRenderingTask(int page, float width, float height, RectF bounds, boolean thumbnail, int cacheOrder, boolean bestQuality, boolean annotationRendering) {
        RenderingTask task = new RenderingTask(width, height, bounds, page, thumbnail, cacheOrder, bestQuality, annotationRendering);
        Message msg = obtainMessage(MSG_RENDER_TASK, task);
        sendMessage(msg);
    }

    @Override
    public void handleMessage(Message message) {
        final RenderingTask task = (RenderingTask) message.obj;
        try {
            final PagePart part = proceed(task, 0);
            if(listener != null) {
                if (part != null) {
                    if (running) {
                        listener.onTaskBitmapRendered(part);
                    } else {
                        listener.onTaskBitmapRecycle(part);
                    }
                } else {
                    listener.onTaskBitmapFailed(task);
                }
            }
        } catch (PageRenderingException ex) {
            ex.printStackTrace();
            if(listener != null) {
                listener.onTaskPageError(ex);
            }
        }
    }

    private PagePart proceed(RenderingTask renderingTask, int retry) throws PageRenderingException {
        PdfFile pdfFile = pdfView.pdfFile;
        if(pdfFile == null || retry > MAX_RETRY){
            throw new PageRenderingException(renderingTask.page, new Throwable("pdfFile is null"));
        }
        pdfFile.openPage(renderingTask.page);

        int w = Math.round(renderingTask.width);
        int h = Math.round(renderingTask.height);

        Log.d(TAG, "proceed: " + renderingTask.cacheOrder + ", w: " + w + ", h: " + h);
        if (w == 0 || h == 0 || pdfFile.pageHasError(renderingTask.page)) {
            return null;
        }

        Bitmap render;
        try {
            Bitmap.Config config = renderingTask.bestQuality ? Bitmap.Config.ARGB_8888 : Bitmap.Config.RGB_565;
            try {
                render = Bitmap.createBitmap(pdfView.getContext().getResources().getDisplayMetrics(), w, h, config);
            }catch (NullPointerException npe){
                Log.e(TAG, "display metrics is null, load bitmap without it", npe);
                render = Bitmap.createBitmap(w, h, config);
            } catch (OutOfMemoryError ome){
                System.gc();
                return proceed(renderingTask, retry + 1);
            }
        } catch (IllegalArgumentException e) {
            Log.e(TAG, "Cannot create bitmap", e);
            return null;
        }
        calculateBounds(w, h, renderingTask.bounds);

        pdfFile.renderPageBitmap(render, renderingTask.page, roundedRenderBounds, renderingTask.annotationRendering, pdfView.isBlackAndWhite());

        return new PagePart(renderingTask.page, render, renderingTask.bounds, renderingTask.thumbnail, renderingTask.cacheOrder);
    }

    private void calculateBounds(int width, int height, RectF pageSliceBounds) {
        renderMatrix.reset();
        renderMatrix.postTranslate(-pageSliceBounds.left * width, -pageSliceBounds.top * height);
        renderMatrix.postScale(1 / pageSliceBounds.width(), 1 / pageSliceBounds.height());

        renderBounds.set(0, 0, width, height);
        renderMatrix.mapRect(renderBounds);
        renderBounds.round(roundedRenderBounds);
    }

    void stop() {
        running = false;
    }

    void start() {
        running = true;
    }

    public OnRenderingHandlerListener getListener() {
        return listener;
    }

    public void setListener(OnRenderingHandlerListener listener) {
        this.listener = listener;
    }

    public class RenderingTask {
        float width, height;
        RectF bounds;
        int page;
        boolean thumbnail;
        int cacheOrder;
        boolean bestQuality;
        boolean annotationRendering;

        RenderingTask(float width, float height, RectF bounds, int page, boolean thumbnail, int cacheOrder, boolean bestQuality, boolean annotationRendering) {
            this.page = page;
            this.width = width;
            this.height = height;
            this.bounds = bounds;
            this.thumbnail = thumbnail;
            this.cacheOrder = cacheOrder;
            this.bestQuality = bestQuality;
            this.annotationRendering = annotationRendering;
        }
    }
}
