//
//  PdfUtility.m
//  RCTPdf
//
//  Created by iPad Developer on 19.09.19.
//  Copyright © 2019 wonday.org. All rights reserved.
//

#import "PdfUtility.h"
#import <UIKit/UIKit.h>
#import <React/RCTLog.h>
#import <React/RCTUIManager.h>
#if __has_include(<React/RCTUIManagerUtils.h>)
#import <React/RCTUIManagerUtils.h>
#endif
#import <React/RCTBridge.h>
#import "RCTPdfBoxView.h"
#import "RCTGraphBoxView.h"
#import "RCTImageView.h"

@implementation PdfUtility

RCT_EXPORT_MODULE();

@synthesize bridge = _bridge;

- (id)init
{
    self = [super init];
    
    if (self) {
        pngQueue = [[NSOperationQueue alloc] init];
        [pngQueue setMaxConcurrentOperationCount:1];
    }
    
    return self;
}

+ (BOOL)requiresMainQueueSetup
{
  return NO;
}

- (dispatch_queue_t)methodQueue
{
    return RCTGetUIManagerQueue();
}

RCT_REMAP_METHOD(pdfToPng, config:(NSDictionary *)config resolve:(RCTPromiseResolveBlock)resolve
                 reject:(RCTPromiseRejectBlock)reject)
{
    if (pngQueue == nil) {
        pngQueue = [[NSOperationQueue alloc] init];
        [pngQueue setMaxConcurrentOperationCount:1];
    } else {
        NSLog(@"pdfToPng, start queue length: %lu", (unsigned long)[pngQueue operationCount]);
    }
    NSBlockOperation *blockOp = [NSBlockOperation blockOperationWithBlock:^{
        @try {
            NSString *source = [config objectForKey:@"source"];
            NSString *destination = [config objectForKey:@"destination"];
            int pageId = [[config objectForKey:@"page"] intValue];
            BOOL isDoubleTruck = [[config objectForKey:@"isDoubleTruck"] boolValue];
            int graphWidth = [[config objectForKey:@"graphWidth"] intValue];
            int graphHeight = [[config objectForKey:@"graphHeight"] intValue];
            double pngQualityFactor = [[config objectForKey:@"pngQualityFactor"] doubleValue];
            int width = [[config objectForKey:@"width"] intValue];
            int height = [[config objectForKey:@"height"] intValue];
            float pixelRatio = [[config objectForKey:@"pixelRatio"] floatValue];
            BOOL generatePNGBasedOnScreenDimensions = [[config objectForKey:@"generatePNGBasedOnScreenDimensions"] boolValue];
            int pngWidthLimit = [[config objectForKey:@"pngWidthLimit"] intValue];
            
            #if DEBUG
                #if (DEBUG==1)
                    RCTLogInfo(@"Generate text PNG from PDF for page %i, %@", pageId, source);
                #endif
            #endif
            //check if the PNG image have been already generated
            if ([[NSFileManager defaultManager] fileExistsAtPath:destination]) {
                #if DEBUG
                    #if (DEBUG==1)
                        RCTLogInfo(@"Text PNG image already present for page %i, %@", pageId, source);
                    #endif
                #endif
                resolve(destination);
                return;
            } else {
                NSString *fallbackPng = [destination stringByReplacingOccurrencesOfString:@"pagetext.png" withString:@"text.png"];
                if ([[NSFileManager defaultManager] fileExistsAtPath:fallbackPng]) {
                   #if DEBUG
                      #if (DEBUG==1)
                        RCTLogInfo(@"Fallback Text PNG image already present for page %i, %@", pageId, source);
                       #endif
                   #endif
                   resolve(fallbackPng);
                   return;
               }
            }
            
            if ([[NSFileManager defaultManager] fileExistsAtPath:source]) {
                //create temporary file to check if the PNG generation has crashed
                NSString *tempCheckFile = [destination stringByAppendingPathExtension:@"proc"];
                if ([[NSFileManager defaultManager] fileExistsAtPath:tempCheckFile]) {
                    //the file exists it means that the app crashed
                    reject(@"Error", [NSString stringWithFormat:@"Unable to generate PNG from PDF at path, potential crash %@", destination], nil);
                    return;
                } else {
                    NSTimeInterval timestamp = [[NSDate date] timeIntervalSince1970];
                    NSString *tempContent = [NSString stringWithFormat:@"%f", timestamp];
                    [tempContent writeToFile:tempCheckFile atomically:YES encoding:NSUTF8StringEncoding error:nil];
                }
                
                //generate the PNG file
                NSURL *pdfURL = [NSURL fileURLWithPath:source isDirectory:NO];
                
                CGPDFDocumentRef pdf = CGPDFDocumentCreateWithURL((CFURLRef)pdfURL);
                
                // Get the PDF Page that we will be drawing
                CGPDFPageRef page = CGPDFDocumentGetPage(pdf, 1);
                
                // determine the size of the PDF page
                CGRect pageRect = CGPDFPageGetBoxRect(page, kCGPDFCropBox);
                
                //check that we have a valid PDF size
                if (CGRectIsEmpty(pageRect)) {
                    reject(@"Error", [NSString stringWithFormat:@"Unable to get PDF size from file %@", source], nil);
                } else {
                    float pdfScale = 1.0;
                    
                    if (generatePNGBasedOnScreenDimensions) {
                        float pngWidth = width*pixelRatio;
                        if (pngWidthLimit > 0 && pngWidth > pngWidthLimit) {
                            pngWidth = pngWidthLimit;
                        }
                        if(isDoubleTruck){
                            pngWidth = pngWidth*2;
                        }
                        
                        pdfScale = pngWidth/pageRect.size.width;
                        pageRect.size = CGSizeMake(pngWidth, pageRect.size.height*pdfScale);
                    } else {
                        pdfScale = fmin(graphWidth*pngQualityFactor/pageRect.size.width,graphHeight*pngQualityFactor/pageRect.size.height);
                        
                        if(isDoubleTruck){
                            pdfScale = pdfScale*2;
                        }
                        
                        pageRect.size = CGSizeMake(pageRect.size.width*pdfScale, pageRect.size.height*pdfScale);
                    }
                    
                    // Create a low res image representation of the PDF page to display before the TiledPDFView
                    // renders its content.
                    UIGraphicsBeginImageContext(pageRect.size);
                    CGContextRef context = UIGraphicsGetCurrentContext();
                    // right side up.
                    CGContextScaleCTM(context, 1.0, -1.0);
                    CGContextTranslateCTM(context, -pageRect.origin.x*pdfScale, -pageRect.size.height - pageRect.origin.y*pdfScale);
                    
                    // Scale the context so that the PDF page is rendered
                    // at the correct size for the zoom level.
                    CGContextScaleCTM(context, pdfScale, pdfScale);
                    CGContextSetInterpolationQuality(context, kCGInterpolationNone); //moved from High to avoid sharpening problem of mid res
                    CGContextSetRenderingIntent(context, kCGRenderingIntentDefault);
                    CGContextDrawPDFPage(context, page);
                    
                    UIImage *image = UIGraphicsGetImageFromCurrentImageContext();
                    
                    UIGraphicsEndImageContext();
                    CGPDFDocumentRelease(pdf);
                    
                    NSData *data = UIImagePNGRepresentation(image);
                    [data writeToFile:destination atomically:YES];
                    
                    if ([[NSFileManager defaultManager] fileExistsAtPath:destination]) {
                        #if DEBUG
                          #if (DEBUG==1)
                            RCTLogInfo(@"Text PNG image generated for page %i, %@", pageId, source);
                          #endif
                        #endif
                        //the generation worked remove tmp file
                        if ([[NSFileManager defaultManager] fileExistsAtPath:tempCheckFile]) {
                            [[NSFileManager defaultManager] removeItemAtPath:tempCheckFile error:nil];
                        }
                        resolve(destination);
                    } else {
                        reject(@"Error", [NSString stringWithFormat:@"Unable to generate PNG from PDF at path %@", destination], nil);
                    }
                }
            } else {
                reject(@"Error", [NSString stringWithFormat:@"No PDF file present at %@", source], nil);
            }
        } @catch (NSException *exception) {
            reject(@"Error", [NSString stringWithFormat:@"Unable to generate PNG from PDF file %@", exception], nil);
        }
    }];
    
    [blockOp setQueuePriority:NSOperationQueuePriorityVeryLow];
    [pngQueue addOperation:blockOp];
    NSLog(@"pdfToPng, queue length: %lu", (unsigned long)[pngQueue operationCount]);
}

+ (NSData *)dataWithBase64EncodedString:(NSString *)string {
    //
    //  NSData+Base64.m
    //
    //  Version 1.0.2
    //
    //  Created by Nick Lockwood on 12/01/2012.
    //  Copyright (C) 2012 Charcoal Design
    //
    //  Distributed under the permissive zlib License
    //  Get the latest version from here:
    //
    //  https://github.com/nicklockwood/Base64
    //
    //  This software is provided 'as-is', without any express or implied
    //  warranty.  In no event will the authors be held liable for any damages
    //  arising from the use of this software.
    //
    //  Permission is granted to anyone to use this software for any purpose,
    //  including commercial applications, and to alter it and redistribute it
    //  freely, subject to the following restrictions:
    //
    //  1. The origin of this software must not be misrepresented; you must not
    //  claim that you wrote the original software. If you use this software
    //  in a product, an acknowledgment in the product documentation would be
    //  appreciated but is not required.
    //
    //  2. Altered source versions must be plainly marked as such, and must not be
    //  misrepresented as being the original software.
    //
    //  3. This notice may not be removed or altered from any source distribution.
    //
    const char lookup[] = {
            99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
            99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
            99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 62, 99, 99, 99, 63,
            52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 99, 99, 99, 99, 99, 99,
            99,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 99, 99, 99, 99, 99,
            99, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
            41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 99, 99, 99, 99, 99
    };

    NSData *inputData = [string dataUsingEncoding:NSASCIIStringEncoding allowLossyConversion:YES];
    long long inputLength = [inputData length];
    const unsigned char *inputBytes = [inputData bytes];

    long long maxOutputLength = (inputLength / 4 + 1) * 3;
    NSMutableData *outputData = [NSMutableData dataWithLength:(NSUInteger)maxOutputLength];
    unsigned char *outputBytes = (unsigned char *)[outputData mutableBytes];

    int accumulator = 0;
    long long outputLength = 0;
    unsigned char accumulated[] = {0, 0, 0, 0};
    for (long long i = 0; i < inputLength; i++) {
        unsigned char decoded = lookup[inputBytes[i] & 0x7F];
        if (decoded != 99) {
            accumulated[accumulator] = decoded;
            if (accumulator == 3) {
                outputBytes[outputLength++] = (accumulated[0] << 2) | (accumulated[1] >> 4);
                outputBytes[outputLength++] = (accumulated[1] << 4) | (accumulated[2] >> 2);
                outputBytes[outputLength++] = (accumulated[2] << 6) | accumulated[3];
            }
            accumulator = (accumulator + 1) % 4;
        }
    }

    //handle left-over data
    if (accumulator > 0) outputBytes[outputLength] = (accumulated[0] << 2) | (accumulated[1] >> 4);
    if (accumulator > 1) outputBytes[++outputLength] = (accumulated[1] << 4) | (accumulated[2] >> 2);
    if (accumulator > 2) outputLength++;

    //truncate data to match actual output length
    outputData.length = (NSUInteger)outputLength;
    return outputLength? outputData: nil;
}

RCT_REMAP_METHOD(print, params:(NSDictionary *)params resolve:(RCTPromiseResolveBlock)resolve
                  reject:(RCTPromiseRejectBlock)reject)
{
    [self.bridge.uiManager addUIBlock:^(__unused RCTUIManager *uiManager, NSDictionary<NSNumber *, UIView *> *viewRegistry) {
        
        NSNumber *target = [params objectForKey:@"target"];
        NSDictionary *pdfRealSize = [params objectForKey:@"size"];
        float pdfWidth = [[pdfRealSize objectForKey:@"width"] floatValue];
        float pdfHeight = [[pdfRealSize objectForKey:@"height"] floatValue];
        float pdfMargin = [[pdfRealSize objectForKey:@"margin"] floatValue];
        float scale = [[params objectForKey:@"scale"] floatValue];
        float pdfQuality = [[params objectForKey:@"pdfQuality"] floatValue];
        
        // Get view
        UIView *view;
        
        if ([target intValue] == -1) {
            UIWindow *window = [[UIApplication sharedApplication] keyWindow];
            view = window.rootViewController.view;
        } else {
            view = viewRegistry[target];
        }
        
        if (!view) {
            reject(RCTErrorUnspecified, [NSString stringWithFormat:@"No view found with reactTag: %@", target], nil);
            return;
        }
        
        // Get options
        CGSize size = view.bounds.size;
        
        if (size.width < 0.1 || size.height < 0.1) {
            reject(RCTErrorUnspecified, [NSString stringWithFormat:@"The content size must not be zero or negative. Got: (%g, %g)", size.width, size.height], nil);
            return;
        }
        NSError *error = nil;
        
        if (error) {
            reject(RCTErrorUnspecified, error.localizedDescription, error);
            return;
        }
        
        CGRect scaledViewRect = CGRectApplyAffineTransform(view.frame, CGAffineTransformScale(CGAffineTransformIdentity, scale, scale));
        
        //consider margins
        pdfWidth -= 2*pdfMargin;
        pdfHeight -= 2*pdfMargin;
        
        float printPageScaleX = 1.0;
        float printPageScaleY = 1.0;
        if (pdfWidth < scaledViewRect.size.width) {
            printPageScaleX = pdfWidth / scaledViewRect.size.width;
        }
        if (pdfHeight < scaledViewRect.size.height) {
            printPageScaleY = pdfHeight / scaledViewRect.size.height;
        }
        
        float printPageScale = MIN(printPageScaleX, printPageScaleY);
        
        float offsetX = 0.0;
        if (pdfWidth > scaledViewRect.size.width * printPageScale) {
            offsetX = (pdfWidth - scaledViewRect.size.width * printPageScale) / 2;
        } else {
            offsetX = pdfMargin;
        }
        float offsetY = 0.0;
        if (pdfHeight > scaledViewRect.size.height * printPageScale) {
            offsetY = (pdfHeight - scaledViewRect.size.height * printPageScale) / 2;
        } else {
            offsetY = pdfMargin;
        }
        
        float totalScale = scale * printPageScale;
        
        UIGraphicsBeginImageContextWithOptions(CGSizeMake(pdfWidth, pdfHeight), NO, 0.0);
        
        NSMutableArray *hrRects = [[NSMutableArray alloc] init];
        //loop on subviews
        
        //HR
        for (UIView *subview in view.subviews) {
            for (UIView *component in subview.subviews) {
                
                CGRect scaledRect = CGRectApplyAffineTransform(subview.frame, CGAffineTransformScale(CGAffineTransformIdentity, totalScale, totalScale));
                scaledRect = CGRectApplyAffineTransform(scaledRect, CGAffineTransformTranslate(CGAffineTransformIdentity, offsetX, offsetY));
                if ([component.subviews count] > 0) {
                    for (id subcomponent in component.subviews) {
                        if ([subcomponent isKindOfClass:[RCTView class]]){
                            NSString *hrImage = [subcomponent accessibilityLabel];
                            //NSLog(@"HR image file: %@", hrImage);
                            if (hrImage) {
                                UIImage *newHrImage = nil;
                                if ([hrImage containsString:@"pag_"] && [[NSFileManager defaultManager] fileExistsAtPath:hrImage]) {
                                    newHrImage = [UIImage imageWithContentsOfFile:hrImage];
                                } else {
                                    //base 64
                                    NSData *data = [PdfUtility dataWithBase64EncodedString:hrImage];
                                    if (data != nil) {
                                        newHrImage = [UIImage imageWithData:data];
                                    }
                                }

                                if (newHrImage != nil) {
                                    UIImageView *newImageView = [[UIImageView alloc] initWithImage:newHrImage];
                                    [newImageView drawViewHierarchyInRect:scaledRect afterScreenUpdates:YES];
                                    [hrRects addObject:[NSValue valueWithCGRect:scaledRect]];
                                }
                            }
                        }
                    }
                } else if ([component isKindOfClass:[RCTView class]]){
                    NSString *hrImage = [component accessibilityLabel];
                    //NSLog(@"HR image file: %@", hrImage);
                    if (hrImage) {
                        UIImage *newHrImage = nil;
                        if ([hrImage containsString:@"pag_"] && [[NSFileManager defaultManager] fileExistsAtPath:hrImage]) {
                            newHrImage = [UIImage imageWithContentsOfFile:hrImage];
                        } else {
                            //base 64
                            NSData *data = [PdfUtility dataWithBase64EncodedString:hrImage];
                            if (data != nil) {
                                newHrImage = [UIImage imageWithData:data];
                            }
                        }

                        if (newHrImage != nil) {
                            UIImageView *newImageView = [[UIImageView alloc] initWithImage:newHrImage];
                            [newImageView drawViewHierarchyInRect:scaledRect afterScreenUpdates:YES];
                            [hrRects addObject:[NSValue valueWithCGRect:scaledRect]];
                        }
                    }
                }
            }
        }
        
        //check if there are missing HR boxes
        for (UIView *subview in view.subviews) {
            for (id component in subview.subviews) {
                
                CGRect scaledRect = CGRectApplyAffineTransform(subview.frame, CGAffineTransformScale(CGAffineTransformIdentity, totalScale, totalScale));
                scaledRect = CGRectApplyAffineTransform(scaledRect, CGAffineTransformTranslate(CGAffineTransformIdentity, offsetX, offsetY));
                
                if ([component isKindOfClass:[RCTGraphBoxView class]]){
                    BOOL missingHR = YES;
                    
                    for (id hrRect in hrRects) {
                        CGRect rect = [hrRect CGRectValue];
                        if (CGRectEqualToRect(rect, scaledRect)) {
                            missingHR = NO;
                        }
                    }
                    
                    if (missingHR) {
                        RCTGraphBoxView *hrImageView = (RCTGraphBoxView *)component;
                        [hrImageView drawViewHierarchyInRect:scaledRect afterScreenUpdates:YES];
                    }
                }
            }
        }
        
        //PDF
        for (UIView *subview in view.subviews) {
           for (id component in subview.subviews) {
               
               CGRect scaledRect = CGRectApplyAffineTransform(subview.frame, CGAffineTransformScale(CGAffineTransformIdentity, totalScale, totalScale));
               scaledRect = CGRectApplyAffineTransform(scaledRect, CGAffineTransformTranslate(CGAffineTransformIdentity, offsetX, offsetY));
               
               if ([component isKindOfClass:[RCTPdfBoxView class]]) {
                   RCTPdfBoxView *pdfboxView = (RCTPdfBoxView *)component;
                   UIImage *image = [pdfboxView cropImageAtScale:pdfQuality];
                   //put in a new UIImageView not to affect the box inside the article graph
                   UIImageView *boxView = [[UIImageView alloc] initWithFrame:scaledRect];                 
                   [boxView setImage:image];
                   [boxView drawViewHierarchyInRect:scaledRect afterScreenUpdates:YES];
               }
           }
       }
        
        UIImage *image = UIGraphicsGetImageFromCurrentImageContext();
        UIGraphicsEndImageContext();
        
        if (!image) {
            reject(RCTErrorUnspecified, @"Failed to capture view snapshot. UIGraphicsGetImageFromCurrentImageContext() returned nil!", nil);
            return;
        }
        
        // Convert image to data (on a background thread)
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
            
            NSData *data = UIImagePNGRepresentation(image);
            
            NSError *error = nil;
            NSString *res = nil;
            
            // Save to a temp file
            NSString *path = RCTTempFilePath(@"png", &error);
            if (path && !error) {
                if ([data writeToFile:path options:(NSDataWritingOptions)0 error:&error]) {
                    res = [NSString stringWithFormat:@"file://%@",path];
                }
            }
            
            if (res && !error) {
                resolve(res);
                return;
            }
            
            // If we reached here, something went wrong
            if (error) reject(RCTErrorUnspecified, error.localizedDescription, error);
            else reject(RCTErrorUnspecified, @"viewshot unknown error", nil);
        });
        
    }];
}

@end
