/**
 * Copyright (c) 2017-present, Wonday (@wonday.org)
 * All rights reserved.
 *
 * This source code is licensed under the MIT-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#import "RCTGraphBoxView.h"

#import <Foundation/Foundation.h>
#import <QuartzCore/QuartzCore.h>
#import "FastRCTiledLayer.h"

#if __has_include(<React/RCTAssert.h>)
#import <React/RCTBridgeModule.h>
#import <React/RCTEventDispatcher.h>
#import <React/UIView+React.h>
#import <React/RCTLog.h>
#else
#import "RCTBridgeModule.h"
#import "RCTEventDispatcher.h"
#import "UIView+React.h"
#import "RCTLog.h"
#endif

#ifndef __OPTIMIZE__
// only output log when debug
#define DLog( s, ... ) NSLog( @"<%p %@:(%d)> %@", self, [[NSString stringWithUTF8String:__FILE__] lastPathComponent], __LINE__, [NSString stringWithFormat:(s), ##__VA_ARGS__] )
#else
#define DLog( s, ... )
#endif

// output log both debug and release
#define RLog( s, ... ) NSLog( @"<%p %@:(%d)> %@", self, [[NSString stringWithUTF8String:__FILE__] lastPathComponent], __LINE__, [NSString stringWithFormat:(s), ##__VA_ARGS__] )

@implementation RCTGraphBoxView
{
    CGImageRef _image;
    BOOL _initialed;
    NSArray<NSString *> *_changedProps;
    CGRect _containerFrame;
}

- (instancetype)init
{
    self = [super init];
    if (self) {
        _image = NULL;
        _coordinates = NULL;
        _pageId = 0;
        _artId = -1;
        
        _initialed = NO;
        _changedProps = NULL;
    }
    
    return self;
}

- (UIImage *) cropImage {
    NSDictionary *boxOrigin = [_coordinates objectForKey:@"origin"];
    NSDictionary *boxSize = [_coordinates objectForKey:@"size"];
    float x = [[boxOrigin objectForKey:@"x"] floatValue];
    float y = [[boxOrigin objectForKey:@"y"] floatValue];
    float width = [[boxSize objectForKey:@"width"] floatValue];
    float height = [[boxSize objectForKey:@"height"] floatValue];
    
    CGRect cropRect = CGRectMake(x, y, width, height);
    
    UIGraphicsBeginImageContext(cropRect.size);
    
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextScaleCTM(context, 1.0, -1.0);
    CGImageRef cutImageRef = CGImageCreateWithImageInRect(_image, cropRect);
    CGContextSaveGState(context);
    CGContextRestoreGState(context);
    
    UIImage* croppedImage = [UIImage imageWithCGImage:cutImageRef];
    
    CGImageRelease(cutImageRef);
    
    UIGraphicsEndImageContext();
    
    return croppedImage;
}

- (CGImageRef)imageFromGifData:(NSData *)data {
    CGImageSourceRef source = CGImageSourceCreateWithData((__bridge CFDataRef)data, NULL);
    if (!source) {
        return nil;
    }

    CGImageRef cgImage = CGImageSourceCreateImageAtIndex(source, 0, NULL);
    CFRelease(source);

    return cgImage;
}

- (void)didSetProps:(NSArray<NSString *> *)changedProps
{
    if (!_initialed) {
        
        _changedProps = changedProps;
        
    } else {
        
        if ([changedProps containsObject:@"path"]) {
            
            NSURL *fileURL = [NSURL fileURLWithPath:_path];
            
            if (_image != NULL) {
                //Release old doc
                CGImageRelease(_image);
                _image = NULL;
            }
            
            if ([_path rangeOfString:@".gif"].location != NSNotFound) {
                NSData* gifOriginalImageData = [NSData dataWithContentsOfURL:fileURL];
                _image = [self imageFromGifData:gifOriginalImageData];
            } else {
                CGDataProviderRef dataProvider = CGDataProviderCreateWithURL((__bridge CFURLRef) fileURL);
                _image = CGImageCreateWithJPEGDataProvider(dataProvider, NULL, YES, kCGRenderingIntentDefault);
            }
        }
        
        if (_image && [changedProps containsObject:@"coordinates"]) {
            
            [self setImage:[self cropImage]];
            if (_onBoxDrawn) {
                _onBoxDrawn(@{});
            }
        }
    }
}


- (void)reactSetFrame:(CGRect)frame
{
    [super reactSetFrame:frame];
    
    _initialed = YES;
    
    [self didSetProps:_changedProps];
}

- (void)dealloc{
    _image = NULL;
}

@end
