/**
 * Copyright (c) 2017-present, Wonday (@wonday.org)
 * All rights reserved.
 *
 * This source code is licensed under the MIT-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#import "RCTPdfBoxView.h"

#import <Foundation/Foundation.h>
#import <QuartzCore/QuartzCore.h>
#import "FastRCTiledLayer.h"

#if __has_include(<React/RCTAssert.h>)
#import <React/RCTBridgeModule.h>
#import <React/RCTEventDispatcher.h>
#import <React/UIView+React.h>
#import <React/RCTLog.h>
#else
#import "RCTBridgeModule.h"
#import "RCTEventDispatcher.h"
#import "UIView+React.h"
#import "RCTLog.h"
#endif

#ifndef __OPTIMIZE__
// only output log when debug
#define DLog( s, ... ) NSLog( @"<%p %@:(%d)> %@", self, [[NSString stringWithUTF8String:__FILE__] lastPathComponent], __LINE__, [NSString stringWithFormat:(s), ##__VA_ARGS__] )
#else
#define DLog( s, ... )
#endif

// output log both debug and release
#define RLog( s, ... ) NSLog( @"<%p %@:(%d)> %@", self, [[NSString stringWithUTF8String:__FILE__] lastPathComponent], __LINE__, [NSString stringWithFormat:(s), ##__VA_ARGS__] )

@implementation RCTPdfBoxView
{
    CGPDFDocumentRef _pdfDocument;
    BOOL _initialed;
    NSArray<NSString *> *_changedProps;
    CGRect _containerFrame;
}

- (instancetype)init
{
    self = [super init];
    if (self) {
        _pdfDocument = NULL;
        _coordinates = NULL;
        _offset = NULL;
        _pageId = 0;
        _artId = -1;
        _scale = 1;
        
        _initialed = NO;
        _changedProps = NULL;
    }
    
    return self;
}

- (UIImage *) cropPDFToImage:(CGRect)cropRect atScale:(float)scale{
    
    float offsetX = 0.0;
    float offsetY = 0.0;
    float offsetScale = scale;
    
    if (_offset != NULL) {
        offsetX = [[_offset objectForKey:@"x"] floatValue];
        offsetY = [[_offset objectForKey:@"y"] floatValue];
        offsetScale = [[_offset objectForKey:@"scale"] floatValue];
    }
    
    float min_dpi = 100.0 / 72;
    float max_dpi = 3.0 * min_dpi;
    
    float qualityScale = scale * min_dpi;
    if (_visible) {
        //avoid to show to high quality when zoomed out
        if (scale < 1.0) {
            qualityScale = min_dpi;
        }
    } else {
        //guarantee a good resolution for areas outside visible one
        if (qualityScale < max_dpi) {
            qualityScale = max_dpi;
        }
    }
    if (qualityScale > 5.0 * max_dpi) {
        //limit max quality
        qualityScale = 5.0 * max_dpi;
    }
    
    CGRect scaledCropRect = CGRectMake(cropRect.origin.x*qualityScale, cropRect.origin.y*qualityScale, cropRect.size.width*qualityScale, cropRect.size.height*qualityScale);
    
    CGPDFPageRef PDFPage = CGPDFDocumentGetPage(_pdfDocument, 1);
    CGRect pageRect = CGPDFPageGetBoxRect(PDFPage, kCGPDFCropBox);
    
    float pageWidth = [[_pageSize objectForKey:@"width"] floatValue];
    float pageHeight = [[_pageSize objectForKey:@"height"] floatValue];
    
    CGSize size = CGSizeMake(pageWidth * qualityScale, pageHeight * qualityScale);
    float pageScale = size.height / pageRect.size.height;
    
    UIGraphicsBeginImageContext(scaledCropRect.size);
    
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSetInterpolationQuality(context, kCGInterpolationHigh);
    CGContextTranslateCTM(context, 0, size.height);
    CGContextTranslateCTM(context, -scaledCropRect.origin.x, -scaledCropRect.origin.y);
    CGContextScaleCTM(context, pageScale, -pageScale);
    CGContextSaveGState(context);
    CGContextDrawPDFPage(context, PDFPage);
    CGContextRestoreGState(context);
    
    UIImage *resultingImage = UIGraphicsGetImageFromCurrentImageContext();
    
    UIGraphicsEndImageContext();
    
    return resultingImage;
}

- (UIImage *) cropImageAtScale:(float)scale{
    NSDictionary *boxOrigin = [_coordinates objectForKey:@"origin"];
    NSDictionary *boxSize = [_coordinates objectForKey:@"size"];
    float x = [[boxOrigin objectForKey:@"x"] floatValue];
    float y = [[boxOrigin objectForKey:@"y"] floatValue];
    float width = [[boxSize objectForKey:@"width"] floatValue];
    float height = [[boxSize objectForKey:@"height"] floatValue];
    
    return [self cropPDFToImage:CGRectMake(x, y, width, height) atScale:scale];
}

- (void)didSetProps:(NSArray<NSString *> *)changedProps
{
    if (!_initialed) {
        
        _changedProps = changedProps;
        
    } else {
        
        if ([changedProps containsObject:@"path"]) {
            
            NSURL *fileURL = [NSURL fileURLWithPath:_path];
            
            if (_pdfDocument != NULL) {
                //Release old doc
                CGPDFDocumentRelease(_pdfDocument);
                _pdfDocument = NULL;
            }
            
            _pdfDocument = CGPDFDocumentCreateWithURL((__bridge CFURLRef)fileURL);
            if (_pdfDocument == NULL) {
                 _onPdfError(@{});
            }
        }
        
        if ([changedProps containsObject:@"pageSize"] || [changedProps containsObject:@"scale"] ) {
            [self setImage:[self cropImageAtScale:_scale]];
            [self setFrame:[self getCoordinatesRect]];
            _onBoxDrawn(@{});
        }
    }
}

- (CGRect) getCoordinatesRect {
    NSDictionary *size = [_coordinates objectForKey:@"size"];
    float width = [[size objectForKey:@"width"] floatValue];
    float height = [[size objectForKey:@"height"] floatValue];
    return CGRectMake(0, 0, width, height);
}

- (void)reactSetFrame:(CGRect)frame
{
    [super reactSetFrame:frame];
    
    _containerFrame = frame;
    _initialed = YES;
    
    [self didSetProps:_changedProps];
}

- (void)dealloc{
    CGPDFDocumentRelease(_pdfDocument);
    _pdfDocument = NULL;
}

@end
