/**
 * Copyright (c) 2017-present, Wonday (@wonday.org)
 * All rights reserved.
 *
 * This source code is licensed under the MIT-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#import "RCTPdfView.h"

#import <Foundation/Foundation.h>
#import <QuartzCore/QuartzCore.h>
#import "FastRCTiledLayer.h"

#if __has_include(<React/RCTAssert.h>)
#import <React/RCTBridgeModule.h>
#import <React/RCTEventDispatcher.h>
#import <React/UIView+React.h>
#import <React/RCTLog.h>
#else
#import "RCTBridgeModule.h"
#import "RCTEventDispatcher.h"
#import "UIView+React.h"
#import "RCTLog.h"
#endif

#ifndef __OPTIMIZE__
// only output log when debug
#define DLog( s, ... ) NSLog( @"<%p %@:(%d)> %@", self, [[NSString stringWithUTF8String:__FILE__] lastPathComponent], __LINE__, [NSString stringWithFormat:(s), ##__VA_ARGS__] )
#else
#define DLog( s, ... )
#endif

// output log both debug and release
#define RLog( s, ... ) NSLog( @"<%p %@:(%d)> %@", self, [[NSString stringWithUTF8String:__FILE__] lastPathComponent], __LINE__, [NSString stringWithFormat:(s), ##__VA_ARGS__] )

@implementation RCTPdfView
{
    CGPDFDocumentRef _pdfDocument;
    CGPDFPageRef _PDFPage;
    CGRect _cropBox;
    BOOL _initialed;
    NSArray<NSString *> *_changedProps;
    BOOL _completed;
    CGRect _containerFrame;
    int _boxes;
    int _prevBoxes;
    NSTimer *_timer;
}

- (instancetype)init
{
    self = [super init];
    if (self) {
        _pdfDocument = NULL;
        _PDFPage = NULL;
        _cropBox = CGRectZero;
        
        _page = 1;
        _scale = 1;
        _offsetX = 0;
        _offsetY = 0;
        _isDoubleTruck = NO;
        
        _initialed = NO;
        _changedProps = NULL;
        _completed = NO;
        _boxes = 0;
        _prevBoxes = 0;
        _timer = nil;
    }
    
    return self;
}

// The layer's class should be CATiledLayer.
+ (Class)layerClass
{
    return [FastRCTiledLayer class];
}

-(void)drawLayer:(CALayer *)layer inContext:(CGContextRef)context {
    @synchronized (self) {
        if (_PDFPage != NULL) {
            float containerWidth = _containerFrame.size.width;
            float containerHeight = _containerFrame.size.height;
            
            float pdfPageWidth = _isDoubleTruck ? _cropBox.size.width/2 : _cropBox.size.width;
            float scale = (containerHeight/_cropBox.size.height);
            
            float diffWidth = MAX((containerWidth - pdfPageWidth*scale)/2.0, 0);
            
            CGContextRetain(context);
            CGContextSaveGState(context);
            
            CGContextSetFillColorWithColor(context, [UIColor clearColor].CGColor);
            CGContextFillRect(context, _containerFrame);
            
            CGContextTranslateCTM(context, 0.0, containerHeight);
            
            CGContextTranslateCTM(context, -_cropBox.origin.x*scale + diffWidth, _cropBox.origin.y*scale - (containerHeight - (_cropBox.size.height)*scale));
            CGContextScaleCTM(context, scale, -scale);
            
            CGContextDrawPDFPage(context, _PDFPage);
            
            CGContextRestoreGState(context);
            CGContextRelease(context);
            
            NSLog(@"END boxes 1: prev %i, current %i",_prevBoxes,_boxes);
            _boxes++;
            if (_boxes > 0 && _prevBoxes == 0) {
                NSLog(@"START timer to check boxes");
                
                dispatch_async(dispatch_get_main_queue(), ^{
                    if (self->_timer != nil) {
                        [self->_timer invalidate];
                        self->_timer = nil;
                    }
                    self->_timer = [NSTimer scheduledTimerWithTimeInterval:0.01f
                                                                    target:self selector:@selector(checkBoxes:)
                                                                  userInfo:nil repeats:NO];
                });
            }
            NSLog(@"END boxes 2: prev %i, current %i",_prevBoxes,_boxes);
        } else {
            NSLog(@"PDFPage is NULL");
        }
    }
}

- (void)checkBoxes:(NSTimer *)timer {
    @synchronized (self) {
        if (_PDFPage != NULL) {
            NSLog(@"Check boxes: prev %i, current %i",_prevBoxes,_boxes);
            
            if (self->_timer != nil) {
                [self->_timer invalidate];
                self->_timer = nil;
            }
            
            if (_boxes>0 && _prevBoxes == _boxes && _completed) {
                NSLog(@"Draw PDF completed");
                _onPdfDrawn(@{});
                return;
            }
            
            if (_boxes>0 && _prevBoxes == _boxes && !_completed) {
                _completed = YES;
            } else {
                _completed = NO;
                _prevBoxes = _boxes;
            }
            
            dispatch_async(dispatch_get_main_queue(), ^{
                self->_timer = [NSTimer scheduledTimerWithTimeInterval:0.01f
                                                                target:self selector:@selector(checkBoxes:)
                                                              userInfo:nil repeats:NO];
            });
        }
    }
}

- (void)didSetProps:(NSArray<NSString *> *)changedProps
{
    if (!_initialed) {
        
        _changedProps = changedProps;
        
    } else {
        
        BOOL redraw = NO;
        
        if ([changedProps containsObject:@"path"]) {
            
            NSURL *fileURL = [NSURL fileURLWithPath:_path];
            
            if (_pdfDocument != NULL) {
                //Release old doc
                CGPDFDocumentRelease(_pdfDocument);
                _pdfDocument = NULL;
            }
            
            _pdfDocument = CGPDFDocumentCreateWithURL((__bridge CFURLRef)fileURL);
            if (_pdfDocument != NULL) {
                if (_PDFPage != NULL) {
                    _PDFPage = NULL;
                }
                _PDFPage = CGPDFDocumentGetPage(_pdfDocument, 1);
                _cropBox = CGPDFPageGetBoxRect(_PDFPage, kCGPDFCropBox);
                
                NSLog(@"changed path");
                redraw = YES;
                
                _onPdfLoaded(@{});
            } else {
                _onPdfError(@{});
            }
            
        }
        
        if ([changedProps containsObject:@"scale"]) {
            NSLog(@"changed scale %f", _scale);
            redraw = YES;
        }
        
        if (redraw) {
            _completed = NO;
            _boxes = 0;
            _prevBoxes = 0;
            [self setNeedsDisplay];
        }
    }
}

- (void)reactSetFrame:(CGRect)frame
{
    [super reactSetFrame:frame];
    
    _containerFrame = frame;
    _initialed = YES;
    
    FastRCTiledLayer *tiledLayer = (FastRCTiledLayer *)[self layer];
    tiledLayer.levelsOfDetail = 4;
    tiledLayer.levelsOfDetailBias = 3;
    //check if Retina
    if ([[UIScreen mainScreen] respondsToSelector:@selector(scale)] == YES && [[UIScreen mainScreen] scale] == 2.00)
    {
        tiledLayer.tileSize = CGSizeMake(2048, 2048);
    } else {
        tiledLayer.tileSize = CGSizeMake(1024, 1024);
    }
     
    [self didSetProps:_changedProps];
}

- (void)dealloc{
    NSLog(@"DEALLOC");
    self.layer.contents=nil;
    self.layer.delegate=nil;
    [self.layer removeFromSuperlayer];
    CGPDFDocumentRelease(_pdfDocument);
    _pdfDocument = NULL;
    _PDFPage = NULL;
}

@end
